/*
 * for the Matrox Marvel G200 and Rainbow Runner-G series
 *
 * This module is an interface to the KS0127 video decoder chip.
 *
 * Copyright (C) 1999  Ryan Drake <stiletto@mediaone.net>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 *
 *****************************************************************************
 */
#ifndef __KERNEL__
#define __KERNEL__
#endif
#ifndef MODULE
#define MODULE
#endif

#include <linux/module.h>
#include <linux/delay.h>
#include <linux/errno.h>
#include <linux/kernel.h>
#include <asm/io.h>

#include "zr36060.h"

#define dprintk     if (debug) printk

#define XGENIOCTRL      0x2a
#define XGENIODATA      0x2b

#define VI_IVSYNC    0x0001 /* input vsync interrupt */
#define VI_CMDCMPL   0x0002 /* command complete interrupt */
#define VI_BLVL      0x0004 /* buffer level interrupt */
#define VI_EOI       0x0008 /* codec end of image interrupt */
#define VS_STALLED   0x1000 /* codec stalled after transfer stopped */

#define	VC_CODECEN	0x0001	/* enable codec interface */
#define	VC_MODE		0x0002	/* codec mode VMI/I33 */
#define	VC_TRIG		0x0004	/* codec command trigger */
#define	VC_DATAIN	0x0008	/* Decompress/Compress */
#define	VC_VMIMODE	0x0010	/* VMI mode A/B */
#define	VC_STOP		0x0020	/* Stop after current field */
#define	VC_TRANSEN	0x0040	/* enable transfer to/from frame buffer */

static int debug = 1; /* insmod parameter */

#define	zr36060_delay(x)	mdelay(x)
/****************************************************************************
* linux kernel module api
****************************************************************************/
#ifdef MODULE
int init_module(void)
#else
int zr36060_init(void)
#endif
{

	return 0;
}

#ifdef MODULE
void cleanup_module(void)
{
}
#endif


static void zr36060_misc(struct zr36060 *zr, int value)
{
	int	codecctl;
	int	misc;

	codecctl = (zr->codecctl_shadow | 0x01) & ~0x04;

	writeb(codecctl, zr->codecctl);
	writeb(codecctl, zr->codecctl);
	
	value &= 0xff;
	misc = ~((value & 0x9f) | (~value & 0x60)) & 0xff;
	zr->codecmis_shadow = value & 0xff;

	writeb(misc, zr->codecctl + 3);

	writeb(zr->codecctl_shadow, zr->codecctl);
}

static void zr36060_and_or_codecctl(struct zr36060 *zr, u16 and, u16 or)
{
	int	codecctl;

	codecctl = (zr->codecctl_shadow & and) | or;
	zr->codecctl_shadow = codecctl;

	writew(codecctl, zr->codecctl);
}

static int
zr36060_execute_command(struct zr36060 *zr)
{
	int	status;
	int	i;

	/* tell the system where memory buffer for zr36060 exists */
	writel((zr->host_interface_base & ~0x03) | 0x01, zr->codecaddr);

	if (zr->codecctl_shadow & VC_TRANSEN) {
		zr36060_and_or_codecctl(zr, ~(VC_TRIG | VC_TRANSEN), 0);

		for(i = 255, status = readb(zr->vstatus); 
		    (status & VS_STALLED) == 0 && i--; 
		    status = readb(zr->vstatus)) ;
	}

	/* remove any pending command complete */
	writeb(VI_CMDCMPL, zr->viclear);

	/* start codec command */
	zr36060_and_or_codecctl(zr, ~0x00, VC_TRIG);

	/* wait for codec command to complete */
	for(i = 255, status = readb(zr->vstatus); 
		(status & VI_CMDCMPL) == 0 && i--; 
		status = readb(zr->vstatus)) ;

	/* Ack the command complete interrupt */
	writeb(VI_CMDCMPL, zr->viclear);

	if ((status & VI_CMDCMPL)) {
		return 0;
	}

	printk("zr36060: no command complete on command %x\n", 
	       readw(zr->command_base));
	return 1;
}

static	int
zr36060_write_register(struct zr36060 *zr, int reg, int data)
{
	writew(0x0201, zr->command_base);
	writew((reg << 8) | (data & 0xff), zr->command_base + 8);

	if (zr36060_execute_command(zr) == 0)
		return 0;

	return 1;
}

static	int
zr36060_read_register(struct zr36060 *zr, int reg)
{
	writew(0x8201 | (reg << 10), zr->command_base);

	if (zr36060_execute_command(zr) == 0) {
		return readb(zr->read_base);
	}

	return -1;
}

static int
zr36060_read_buffer(struct zr36060 *zr, int start_reg, char *buffer, int length)
{
	int	data;
	int	i;

	for(i=0; i < length; i++) {
		zr36060_write_register(zr, 1, (start_reg + i) >> 8);
		zr36060_write_register(zr, 2, (start_reg + i));
	
		data = zr36060_read_register(zr, 3);
		if (data < 0) {
			return -1;
		}
		*buffer++ = data;
	}
	return 0;
}
		
static int
zr36060_write_buffer(struct zr36060 *zr, int start_reg, char *buffer, int length)
{
	u8	*host_ptr;
	int	i;

	if (length > 85) {
		printk("zr36060: Write buffer to large %d starting reg %x\n",
		       length, start_reg);
		return 1;
       }

	writeb(0x00, zr->codecctl);
	writeb(0x03, zr->codecctl);
	writeb(0x03, zr->codecctl);

	host_ptr = zr->command_base;

	*host_ptr++ = length * 6;
	*host_ptr++ = 0x02;
	*host_ptr++ = 0;
	*host_ptr++ = 0;
	*host_ptr++ = 0;
	*host_ptr++ = 0;
	*host_ptr++ = 0;
	*host_ptr++ = 0;

	for(i=0; i < length; i++) {
		*host_ptr++ = 1;
		*host_ptr++ = start_reg >> 8;
		*host_ptr++ = 2;
		*host_ptr++ = start_reg;
		*host_ptr++ = 3;
		*host_ptr++ = *buffer++;
	}

	writeb(0x05, zr->codecctl);

	for(i=0; i < 0x7ffff; i++) {
		if ( readl(zr->vstatus) & 0x02) {
printk("command completed\n");
			writeb(0x02, zr->viclear);
			return 0;
		}
	}
printk("timeout codec status register = %x\n", readl(zr->vstatus));

	return 1;
}

static u8 SOF[19] = {
	0xff, 0xc0,		/* Start of frame marker */
	0x00, 0x11,		/* Length of this segment */
	0x08,			/* Precision */
	240/8, 240&0xff,	/* number of lines */
	704/8, 704&0xff,	/* line length */
	3,			/* number of components */
	0, 0x21, 0,		/* Y component */
	1, 0x11, 1,		/* U component */
	2, 0x11, 1		/* V component */
};

static u8 SOS[] = {
	0xff, 0xda,		/* start of Scan marker */
	0x00, 0x0c,		/* lenght of this segment */
	0x03,			/* number of components */
	0x00, 0x00,		/* Y component */
	0x01, 0x11,		/* U component */
	0x02, 0x11,		/* V component */
	0x00, 0x3f, 0x00	/* constant */
};

static u8 DRI[] = {
	0xff, 0xdd,		/* define Restart Interval */
	0x00, 0x04,		/* length of this segment */
	0x00, 0x08,		/* lenght of restart interval in mcu units */
};

static u8 DQT[] = {
	0xff, 0xdb,		/* DHT marker */
	0x00, 0x84,		/* DHT length */
	0x00,			/* table ID 0 */
	0x10, 0x0b, 0x0c, 0x0e, 0x0c, 0x0a, 0x10, 0x0e,
	0x0d, 0x0e, 0x12, 0x11, 0x10, 0x13, 0x18, 0x28,
	0x1a, 0x18, 0x16, 0x16, 0x18, 0x31, 0x23, 0x25,
	0x1d, 0x28, 0x3a, 0x33, 0x3d, 0x3c, 0x39, 0x33,
	0x38, 0x37, 0x40, 0x48, 0x5c, 0x4e, 0x40, 0x44,
	0x57, 0x45, 0x37, 0x38, 0x50, 0x6d, 0x51, 0x57,
	0x5f, 0x62, 0x67, 0x68, 0x67, 0x3e, 0x4d, 0x71,
	0x79, 0x70, 0x64, 0x78, 0x5c, 0x65, 0x67, 0x63,
};

static u8 DQT_2[] = {
	0x01,			/* table ID 1 */
	0x11, 0x12, 0x12, 0x18, 0x15, 0x18, 0x2f, 0x1a,
	0x1a, 0x2f, 0x63, 0x42, 0x38, 0x42, 0x63, 0x63,
	0x63, 0x63, 0x63, 0x63, 0x63, 0x63, 0x63, 0x63,
	0x63, 0x63, 0x63, 0x63, 0x63, 0x63, 0x63, 0x63,
	0x63, 0x63, 0x63, 0x63, 0x63, 0x63, 0x63, 0x63,
	0x63, 0x63, 0x63, 0x63, 0x63, 0x63, 0x63, 0x63,
	0x63, 0x63, 0x63, 0x63, 0x63, 0x63, 0x63, 0x63,
	0x63, 0x63, 0x63, 0x63, 0x63, 0x63, 0x63, 0x63
};

static u8 DHT[] = {
      0xff, 0xdb,		/* DHT marker */
      0x00, 0x84,		/* DHT length */
      0x00,			/* table ID 0 */
      0x10, 0x0b, 0x0c, 0x0e, 0x0c, 0x0a, 0x10, 0x0e,
      0x0d, 0x0e, 0x12, 0x11, 0x10, 0x13, 0x18, 0x28,
      0x1a, 0x18, 0x16, 0x16, 0x18, 0x31, 0x23, 0x25,
      0x1d, 0x28, 0x3a, 0x33, 0x3d, 0x3c, 0x39, 0x33,
      0x38, 0x37, 0x40, 0x48, 0x5c, 0x4e, 0x40, 0x44,
      0x57, 0x45, 0x37, 0x38, 0x50, 0x6d, 0x51, 0x57,
      0x5f, 0x62, 0x67, 0x68, 0x67, 0x3e, 0x4d, 0x71,
      0x79, 0x70, 0x64, 0x78, 0x5c, 0x65, 0x67, 0x63,
};

static u8 DHT_2[] = {
      0x01,			/* table ID 1 */
      0x11, 0x12, 0x12, 0x18, 0x15, 0x18, 0x2f, 0x1a,
      0x1a, 0x2f, 0x63, 0x42, 0x38, 0x42, 0x63, 0x63,
      0x63, 0x63, 0x63, 0x63, 0x63, 0x63, 0x63, 0x63,
      0x63, 0x63, 0x63, 0x63, 0x63, 0x63, 0x63, 0x63,
      0x63, 0x63, 0x63, 0x63, 0x63, 0x63, 0x63, 0x63,
      0x63, 0x63, 0x63, 0x63, 0x63, 0x63, 0x63, 0x63,
      0x63, 0x63, 0x63, 0x63, 0x63, 0x63, 0x63, 0x63,
      0x63, 0x63, 0x63, 0x63, 0x63, 0x63, 0x63, 0x63
};

static void
zr36060_segments(struct zr36060 *zr, int width, int height)
{
	u8	sof[19];

	/* set jpg mode */

	/* set video mode */

	memcpy(sof, SOF, sizeof(sof));
	sof[5] = width >> 8;
	sof[6] = width;

	sof[7] = height >> 8;
	sof[8] = height;

	zr36060_write_buffer(zr, 0x060, sof, sizeof(sof));
	zr36060_write_buffer(zr, 0x07a, SOS, sizeof(SOS));
	zr36060_write_buffer(zr, 0x0c0, DRI, sizeof(DRI));
	zr36060_write_buffer(zr, 0x0cc, DQT, sizeof(DQT));
	zr36060_write_buffer(zr, 0x0cc+sizeof(DQT), DQT_2, sizeof(DQT_2));
	zr36060_write_buffer(zr, 0x1d4, DHT, sizeof(DHT));
	zr36060_write_buffer(zr, 0x1d4+sizeof(DHT), DHT_2, sizeof(DHT_2));
/*
	zr36060_write_buffer(zr, 0x380, APP, sizeof(APP));
	zr36060_write_buffer(zr, 0x3c0, COM, sizeof(COM));
*/
}

void	zr36060_2A36(struct zr36060 *zr);
void	zr36060_AA17(struct zr36060 *zr);
void	zr36060_EC5B(struct zr36060 *zr, int xc, int x10);
int	zr36060_F416(struct zr36060 *zr, int xc);
void	zr36060_FB14(struct zr36060 *zr, int xc, int x10);
void	zr36060_FB7C(struct zr36060 *zr);
void	zr36060_FC18(struct zr36060 *zr);
void	zr36060_FC99(struct zr36060 *zr);
void	zr36060_108D8(struct zr36060 *zr, int xc);

void
zr36060_2A36(struct zr36060 *zr)
{
	zr36060_AA17(zr);
	zr36060_FC18(zr);

	zr36060_108D8(zr, 0);
}

void
zr36060_AA17(struct zr36060 *zr)
{
	int	temp_0c;
	int	temp_10;

	temp_0c = zr->_28;

	if (temp_0c != 0) {
		if (temp_0c < 0 || (temp_0c > 2)) {
			zr36060_EC5B(zr, 0x0f, 0);
		} else {
			zr36060_EC5B(zr, 0x05, 0);
		}
	}

	zr36060_EC5B(zr, 0x10, 0);

	zr->_71 = 0;
	zr->_6d = zr->_10;

#ifdef	NEVER
	temp_08 = zr->codecctl_shadow;
	temp_08 &= ~4;
	temp_08 &= ~20;
	temp_08 &= ~40;
	temp_08 &= ~01;
	zr36060_codecctl(zr, temp_08);
#else 
	zr36060_and_or_codecctl(zr, ~0x65, 0);
#endif
	zr36060_delay(100);

#ifdef	NEVER
	temp_08 != 0x01;
	zr36060_codecctl(zr, temp_08);
	zr36060_codecctl(zr, temp_08);
#else
	zr36060_and_or_codecctl(zr, ~0x00, 0x01);
	zr36060_and_or_codecctl(zr, ~0x00, 0x01);
#endif


	temp_10 = zr->_28;

	if (temp_10 != 0) {
		if (temp_10 < 0 || (temp_10 > 2)) {
			zr36060_EC5B(zr, 0x0f, 1);
		} else {
			zr36060_EC5B(zr, 0x05, 1);
		}
	}

	zr36060_EC5B(zr, 0x10, 1);
}

static int lcod_2ee4;
int
zr36060_CC90(struct zr36060 *zr, int reg, int data)
{
	int	temp_04;
	int	temp_08;
	int	temp_10;
	int	temp_14;
	int	temp_16;
	int	temp_18;
	int	temp_1c;
	int	temp_20;
	int	temp_28;

	temp_10 = 0;

	temp_18 = (~reg & 0x07) | (reg & ~0x7);

	temp_28 = 0x0201;

	temp_04 = data & 0xff;
	if (zr->_4 != 0) {
		temp_04 |= ((temp_18 & 0x0f) << 8);
	} else {
		temp_04 |= ((reg & 0x0f) << 8);
	}

	writew(temp_28, zr->_14);
	writew(temp_04, zr->_14+8);

	temp_20 = zr->codecctl_shadow;
	if (((temp_20 >> 6) & 1) == 1) {
		zr36060_and_or_codecctl(zr, ~0x44, 0);
		lcod_2ee4 = 1;
	}

	if (lcod_2ee4) {
		if (((temp_20 >> 1) & 0x01)==0 && ((temp_20 >> 4) & 0x01)==0) {
			temp_1c = 0xff;
			temp_08 = readw(zr->vstatus);
			while(((temp_08 >> 12) & 0x01) == 0 && --temp_1c) {
				temp_08 = readw(zr->vstatus);
			}

			if (temp_1c < 1)
				return 1;
		}
	} else {
		if (zr->_169 == 0 && zr->_28 != 0) {
			zr36060_AA17(zr);
		}
	}

	if (temp_10 == 0) {
		if (zr->_28 == 0) {
/*
			pushd
			cli
*/
		}

		writeb(zr->viclear, 02);

		temp_14 = zr->_10;
		zr36060_and_or_codecctl(zr, ~0, 4);

		temp_16 = 0xff;
		temp_08 = readb(zr->vstatus) & 0xff;

		while(((temp_08 >> 1) & 1) == 0 && temp_16--) {
			temp_08 = readb(zr->vstatus) & 0xff;
		}
		writeb(zr->viclear, 02);

		if (zr->_28 == 0) {
/*			popd */
		}
	}

	if (((temp_20 >> 6) & 1) == 1) {
		zr36060_and_or_codecctl(zr, ~0x04, 0x40);
		lcod_2ee4 = 0;
		return temp_10;
	}

	if (zr->_169 == 0 && zr->_28) {
		zr36060_AA17(zr);
	}

	return temp_10;
}


void
zr36060_EC5B(struct zr36060 *zr, int xc, int x10)
{
	int	temp_48;
	struct	zr36060 *temp_4c;

	temp_4c = zr;
	temp_48 = 0;
	
		

	if (x10 == 0) {
		if (xc & 0x01)
			temp_4c->_45 |= 02;

		if (xc & 0x02)
			temp_4c->_3d |= 2;

		if (xc & 0x04)
			temp_4c->_41 |= 2;

		if (xc & 0x08)
			temp_4c->_49 |= 2;

		if (xc & 0x10)
			temp_4c->_4d |= 2;

	} else {
		if (xc & 0x01)
			temp_4c->_45 &= 0xfd;

		if (xc & 0x02)
			temp_4c->_3d &= ~2;

		if (xc & 0x04)
			temp_4c->_41 &= ~2;

		if (xc & 0x08)
			temp_4c->_49 &= 0xfd;

		if (xc & 0x10)
			temp_4c->_4d &= ~2;
	}

}

int
zr36060_F416(struct zr36060 *zr, int xc)
{
	int	rc;

	rc = 0;

	switch(xc) {

	case 7:
		rc = readb(zr->vstatus) & 0xff;
		break;

	}

	return rc;
		
}


void
zr36060_FB14(struct zr36060 *zr, int xc, int x10)
{
	if (xc & 0xff)
		zr36060_misc(zr, zr->codecmis_shadow & ~0x10);
	else 	
		zr36060_misc(zr, zr->codecmis_shadow | 0x10);
}

void
zr36060_FB7C(struct zr36060 *zr)
{
	int	misc;

	misc = zr->codecmis_shadow;
	zr36060_delay(10);
	zr36060_misc(zr, zr->codecmis_shadow & ~0x04);
	zr36060_delay(1);
	zr36060_misc(zr, zr->codecmis_shadow | 0x04);

	zr36060_misc(zr, misc);
}
	

void
zr36060_FC18(struct zr36060 *zr)
{
	zr36060_FC99(zr);

	zr36060_FB7C(zr);

	zr36060_FC99(zr);
	zr36060_FC99(zr);
	zr36060_FC99(zr);
	zr36060_FC99(zr);
}

void
zr36060_FC99(struct zr36060 *zr)
{
	zr36060_FB14(zr, 0, 0);
	zr36060_delay(100);
	zr36060_misc(zr, zr->codecmis_shadow & ~0x08);
	zr36060_delay(1);
	zr36060_misc(zr, zr->codecmis_shadow | 0x08);
	zr36060_delay(200);
}

void
zr36060_108D8(struct zr36060 *zr, int xc)
{
}


int
zr36060_attach(struct zr36060 *zr)
{
	char	buffer[4];

	/* reset the codec interface engine */
	zr36060_and_or_codecctl(zr, 0x00, 0x00);
	zr36060_and_or_codecctl(zr, 0x00, 0x03);
	zr36060_and_or_codecctl(zr, 0x00, 0x03);

	zr36060_misc(zr, 0x98);
	zr36060_FC18(zr);
	zr36060_2A36(zr);

	/* tell the system where memory buffer for zr36060 exists */
	writel((zr->host_interface_base & ~0x03) | 0x01, zr->codecaddr);

	/* read id from zr36060 */
	if (zr36060_read_buffer(zr, 0x22, buffer, 2) < 0) {
		printk("zr36060: Read error\n");
		return -1;
	}
	if (buffer[0] == 0x33) {
		printk("zr36060: zr36060 found revision %d\n", buffer[1]);
	} else {
		printk("zr36060: zr36060 not found (%x %x)\n",
		       buffer[0], buffer[1]);
		return -1;
	}
	
	
#ifdef	NEVER
	zr36060_segments(zr, 704, 240);
	MOD_INC_USE_COUNT;
#endif

	return 0;
}

int
zr36060_deattach(struct zr36060 *zr)
{
#ifdef	NEVER
	/* reset the codec interface */
	writeb(0x00, zr->codecctl);
	writeb(0x01, zr->codecctl);
	writeb(0x00, zr->codecctl);

	MOD_DEC_USE_COUNT;
#endif
	return 0;
}
